require "openshift-origin-common/utils/path_utils"
require "openshift-origin-common/config"
require "openshift-origin-node/utils/environ"
require "openshift-origin-node/utils/selinux"
require "fileutils"

module OpenShift
  class GearUpgradeExtension

    # When renaming cartridge's versions...
    #
    # Hash key(String) => value(String)
    # cartridge name + '-' + old version => new version
    VERSION_MAP = {
    }

    def self.version
      '2.1'
    end

    def initialize(upgrader)
      @upgrader  = upgrader
      @uuid      = upgrader.uuid
      @gear_home = upgrader.gear_home
      @container = upgrader.container
    end

    def pre_upgrade(progress)
      fix_inode_quota(progress)
      fix_deployment_ownership(progress)
      fix_ownership(progress)
      fix_deployment_metadata_ownership(progress)
    end

    def fix_ownership(progress)
      progress.step 'fix_ownership' do |context, errors|

        begin
          @container.set_ro_permission(@container.container_dir)
        rescue Exception => e
          errors << "Couldn't set ownership of #{@container.container_dir}"
        end

        files = []
        @container.cartridge_model.each_cartridge do |cart|
          files << PathUtils.join(@gear_home, cart.name, "metadata", "manifest.yml")
          files << PathUtils.join(@gear_home, cart.name, "metadata", "managed_files.yml")
          files << PathUtils.join(@gear_home, cart.name, "env", "OPENSHIFT_#{cart.short_name}_IDENT")
          files << PathUtils.join(@gear_home, cart.name, "env", "OPENSHIFT_#{cart.short_name}_DIR")
        end

        files.each do |file|
          next unless File.exists?(file)

          begin
            @container.set_ro_permission(file)
          rescue Exception => e
            errors << "Couldn't set ownership of #{file}"
          end
        end
      end
    end

    def fix_deployment_ownership(progress)
      progress.step 'fix_deployment_ownership' do |context, errors|
        deps_dir = PathUtils.join(@gear_home, "app-root", "runtime", "dependencies")
        build_deps_dir = PathUtils.join(@gear_home, "app-root", "runtime", "build-dependencies")
        mcs_label = OpenShift::Runtime::Utils::SELinux::get_mcs_label(@container.uuid)

        [deps_dir, build_deps_dir].each do |dir|
          next unless Dir.exists?(dir)

          begin
            PathUtils.oo_chown_R(@container.uid, @container.gid, dir)
          rescue Exception => e
            errors << "Couldn't set ownership of #{dir}: #{e.message}"
          end

          begin
            OpenShift::Runtime::Utils::SELinux.set_mcs_label_R(mcs_label, dir)
          rescue Exception => e
            errors << "Couldn't set mcs label '#{mcs_label}' on #{dir}: #{e.message}"
          end
        end
      end
    end 

    # See 09-fix-quota-files. For OSE we are not forcing customers to move to
    # our new default 80k inode setting.  If they had the original default we
    # move them to the new default.  If they had a custom setting this logic is
    # still useful for sanity checking.
    def fix_inode_quota(progress)
      progress.log 'Getting default values from resource_limits.conf'
      # get defaults from resource_limits.conf
      defaults = OpenShift::Runtime::Node.resource_limits

      default_blocks_limit = defaults.get('quota_blocks').to_i
      progress.log "default_blocks_limit = #{default_blocks_limit}"

      default_inodes_limit = defaults.get('quota_files').to_i
      progress.log "default_inodes_limit = #{default_inodes_limit}"

      progress.log "Getting gear quota information"
      # get the current quota for the gear
      quota = OpenShift::Runtime::Node.get_quota(@uuid)

      gear_blocks_limit = quota[:blocks_limit]
      progress.log "original gear_blocks_limit = #{gear_blocks_limit}"

      # make sure we aren't below the default
      gear_blocks_limit = default_blocks_limit if gear_blocks_limit < default_blocks_limit
      progress.log "revised gear_blocks_limit = #{gear_blocks_limit}"

      # this should always be >= 1 because gear_blocks_limit can never be < default_blocks_limit
      multiplier = gear_blocks_limit / default_blocks_limit
      progress.log "Multiplier = #{multiplier}"

      # e.g. if multiplier is 2 and the default inodes limit is 80000, the min should be 160000
      minimum_inodes_limit = default_inodes_limit * multiplier
      progress.log "minimum_inodes_limit = #{minimum_inodes_limit}"

      gear_inodes_limit = quota[:inodes_limit]
      progress.log "gear_inodes_limit = #{gear_inodes_limit}"

      # if the gear's current inodes limit is < the calculated minimum_inodes_limit, increase it
      if gear_inodes_limit < minimum_inodes_limit
        progress.log "Setting quota to #{gear_blocks_limit} blocks, #{minimum_inodes_limit} inodes"
        OpenShift::Runtime::Node.set_quota(@uuid, gear_blocks_limit, minimum_inodes_limit)
      end
    end

    def fix_deployment_metadata_ownership(progress)
      progress.step 'fix_deployment_metadata_ownership' do |context, errors|
        Dir.glob(PathUtils.join(@gear_home,"app-deployments","*","metadata.json")).each do |f|
          begin
            @container.set_rw_permission(f)
          rescue Exception => e
            errors << "Couldn't set ownership of file #{f}"
          end
        end
      end
    end

    unless VERSION_MAP.empty?
      def map_ident(progress, ident)
        vendor, name, version, cartridge_version = OpenShift::Runtime::Manifest.parse_ident(ident)
        name_version                             = "#{name}-#{version}"

        if VERSION_MAP[name_version]
          progress.log "Mapping version #{version} to #{VERSION_MAP[name_version]} for cartridge #{name}"
          version = VERSION_MAP[name_version] || version
        end

        return vendor, name, version, cartridge_version
      end
    end

    def pre_cartridge_upgrade(progress, itinerary)
      create_log_dir(progress)
    end

    def create_log_dir(progress)
      progress.step 'create_global_log_dir' do |context, errors|
        begin
          gearlogdir = PathUtils.join(@gear_home, "app-root") + "/logs/"
          @container.add_env_var("LOG_DIR", gearlogdir, true) {|v|
            FileUtils.mkdir_p(v, :verbose => @debug)
          }
          @container.set_rw_permission_R(gearlogdir)
        end
      end
    end

    def post_cartridge_upgrade(progress, itinerary)
    end

    def migrate_deployment_system_pre(progress)
    end

    def migrate_deployment_system_post(progress)
    end

  end
end
